package com.ipsymcon;

import java.net.Socket;
import java.net.UnknownHostException;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.util.logging.Logger;
import java.util.logging.Level;
import java.util.List;
import java.util.ArrayList;
import java.security.MessageDigest;

/**
 * @author Kosch
 * todo extra timed thread sending PING to avoid timeout for long running clients
 * todo different exception classes for different exceptions
 * todo support for the other fuctions (I have no use-case for them ATM)
 * todo fix maybe character-conversion problems, if source is executed on different plattforms
 */
public class IPSClient {

    private static final Logger LOG=Logger.getLogger(IPSClient.class.getName());

    private String host;
    private Integer port;

    private Socket socket;
    private InputStream is;
    private OutputStream os;
    private int requestedSID;
    protected int sessionID=-1;
    private boolean loggedIn;

    private static final String CMD_PING="PING;";
    private static final String CMD_AUTH="AUTH;";
    private static final String CMD_DOAUTH="DoAuth;";
    private static final String CMD_LISTVARS="ListVars;";
    private static final String CMD_GETVALUE="GetValue;";
    private static final String CMD_SETVALUE="SetValue;";

    public IPSClient(String host, Integer port){
        this.host=host;
        this.port=port;
    }

    public Socket getSocket(){
        return socket;
    }

    public void logon(String user, String password) throws UnknownHostException, IOException, Exception {
        LOG.log(Level.INFO, "Logging in...");
        socket=new Socket(host, port);
        socket.setSoTimeout(10000);
        is=socket.getInputStream();
        os=socket.getOutputStream();

        sessionID=-1;
        requestedSID++;
        String result = sendCommand(CMD_AUTH);

        String key=result.substring(7, result.length()-1);
        LOG.log(Level.INFO, "Key "+key);

        String pwdXOR=XOR(password, key);
        String md5=getMD5(pwdXOR);

        sendCommand(CMD_DOAUTH+user+';'+md5.toLowerCase()+';');
        loggedIn=true;
        LOG.log(Level.INFO, "Logged in.");
    }

    /**
     * Can be used for keep-alive requests.
     * @throws Exception
     */
    public void pingServer() throws Exception {
        if(!loggedIn){
            throw new Exception("Not logged in.");
        }
        sendCommand(CMD_PING);
    }

    /**
     * Get the list of available variables and their types.
     * @return list of Var-Objects without value
     * @throws Exception
     */
    public List<Var> listVariables() throws Exception {
        if(!loggedIn){
            throw new Exception("Not logged in.");
        }
        String result = sendCommand(CMD_LISTVARS);
        String[] vars=result.split(";");
        List<Var> list=new ArrayList<Var>();
        // skip 1st one (its status flag)
        for(int i=1, size=vars.length; i<size; i++){
            int delimiter=vars[i].indexOf('|');
            if(delimiter==-1){
                LOG.log(Level.INFO, "Invalid var description: "+vars[i]);
                continue;
            }
            String name=vars[i].substring(0, delimiter);
            String type=vars[i].substring(delimiter+1);
            Var var=new Var(name);
            if("boolean".equalsIgnoreCase(type)){
                var.setType(Var.Type.BOOLEAN);
            } else if("float".equalsIgnoreCase(type)){
                var.setType(Var.Type.FLOAT);
            } else {
                LOG.log(Level.INFO, "Unknown type: "+type);
                var.setType(Var.Type.UNKNOWN);
            }
            list.add(var);
        }
        return list;
    }

    /**
     * Applies the value from server to the Var.
     * @param var
     * @throws Exception
     */
    public void getValue(Var var) throws Exception {
        if(!loggedIn){
            throw new Exception("Not logged in.");
        }
        String cmd=CMD_GETVALUE+var.getName()+';';
        // bad design, why not send the datatype together with the value?
        if(Var.Type.BOOLEAN==var.getType()){
            cmd+="Boolean;";
        } else if(Var.Type.FLOAT==var.getType()){
            cmd+="Float;";
        } else {
            LOG.log(Level.INFO, "Unknown var type.");
            return;
        }
        String result = sendCommand(cmd);
        String[] parts=result.split(";");
        // 2nd part contains value
        var.setValue(parts[1]);
    }

    /**
     * Applies the value of the Var to the server.
     * @param var
     * @throws Exception
     */
    public void setValue(Var var) throws Exception {
        if(!loggedIn){
            throw new Exception("Not logged in.");
        }
        String cmd=CMD_SETVALUE+var.getName()+';';
        if(Var.Type.BOOLEAN==var.getType()){
            cmd+="Boolean;";
        } else if(Var.Type.FLOAT==var.getType()){
            cmd+="Float;";
        } else {
            LOG.log(Level.INFO, "Unknown var type.");
            return;
        }
        // apply the value
        if(Var.Type.BOOLEAN==var.getType()){
            cmd+=var.booleanValue().toString().toUpperCase();
        } else if(Var.Type.FLOAT==var.getType()){
            String val=var.floatValue().toString();
            val=val.replace('.', ',');
            cmd+=val+';';
        }
        sendCommand(cmd);
    }

    /**
     * Always log off. This closes Socket.
     * @throws IOException
     */
    public void logoff() throws IOException {
        if(socket!=null){
            socket.close();
            loggedIn=false;
        }
    }

    private synchronized String sendCommand(String command) throws IOException, Exception {
        byte[] message=new byte[26+command.length()];
        message[0]='*';
        message[1]='|';
        message[2]='M';
        byte[] int32=getBytes(command.length());
        message[3]=int32[0];
        message[4]=int32[1];
        message[5]=int32[2];
        message[6]=int32[3];
        message[7]='|';
        message[8]='S';
        if(sessionID==-1){
            int32=getBytes(requestedSID);
        } else {
            int32=getBytes(sessionID);
        }
        message[9]=int32[0];
        message[10]=int32[1];
        message[11]=int32[2];
        message[12]=int32[3];
        message[13]='|';
        message[14]='P';
        int32=getBytes(1);
        message[15]=int32[0];
        message[16]=int32[1];
        message[17]=int32[2];
        message[18]=int32[3];
        message[19]='|';
        message[20]='N';
        int32=getBytes(1);
        message[21]=int32[0];
        message[22]=int32[1];
        message[23]=int32[2];
        message[24]=int32[3];
        message[25]='|';
        for(int i=26, size=26+command.length(); i < size; i++){
            message[i]=(byte)command.charAt(i-26);
        }
        LOG.log(Level.INFO, "Send message: "+new String(message));
        os.write(message);
        os.flush();

        // fetch result
        int start=is.read();
        if('*'!=start){
            throw new IOException("Communication async!");
        }
        byte[] tmp=new byte[2];
        is.read(tmp);
        int dataLength=readInt32(is);
        tmp=new byte[2];
        is.read(tmp);
        int assignedSID=readInt32(is);
        if(sessionID==-1){
            sessionID=assignedSID;
        }
        tmp=new byte[13];
        is.read(tmp);
        tmp=new byte[dataLength];
        is.read(tmp);
        String res=new String(tmp);
        LOG.log(Level.INFO, "Read data: "+res);
        if(res.startsWith("FAILED")){
            throw new Exception("Command execution failed: "+res);
        }
        return res;
    }

    private static int readInt32( InputStream is) throws IOException {
        byte[] a=new byte[4];
        is.read(a);
        return a[3] << 24 & 0xff000000
                | a[2] << 16 & 0x00ff0000
                | a[1] << 8 & 0x0000ff00
                | a[0] & 0x000000ff;
    }

    private static byte[] getBytes(int integer) {
        int byteNum = 40 - Integer.numberOfLeadingZeros(integer < 0 ? ~integer : integer) >> 3;
        byte[] byteArray = new byte[4];

        for (int n = 0; n < byteNum; n++)
            byteArray[3 - n] = (byte) (integer >>> (n << 3));

        byte[] revArray=new byte[4];
        revArray[0]=byteArray[3];
        revArray[1]=byteArray[2];
        revArray[2]=byteArray[1];
        revArray[3]=byteArray[0];
        return revArray;
    }

    private static String XOR(String password, String key){
        int KeyInt = Integer.valueOf(key);
        StringBuffer buf=new StringBuffer();
        for (int i = 0; i < password.length(); i++){
            int passwordXOR = password.charAt(i) ^ KeyInt;
            buf.append((char)passwordXOR);
        }
        return buf.toString();
    }

    private static String getMD5(String str){
        try {
            MessageDigest md = MessageDigest.getInstance("MD5");
            md.reset();
            md.update(str.getBytes());
            byte[] hash = md.digest();
            StringBuffer md5=new StringBuffer();
            for(byte b : hash) {
                String hex=Integer.toHexString(0xFF & b);
                if(hex.length()==1){
                    hex='0'+hex;
                }
                md5.append(hex);
            }
            return md5.toString();
        } catch (Exception ex){
            // todo handling
        }
        return "";
    }

    public static class Var {

        public enum Type { BOOLEAN, FLOAT, UNKNOWN }

        private String name;
        private Type type;
        private Object value;

        public Var(String name){
            this.name=name;
        }

        public Type getType() {
            return type;
        }

        public void setType(Type type) {
            this.type = type;
        }

        public String getName() {
            return name;
        }

        public void setValue(Object value){
            this.value=value;
        }

        public Object getValue(){
            return value;
        }

        public Boolean booleanValue(){
            if(value==null){
                return false;
            } else if(value instanceof Boolean){
                return (Boolean)value;
            } else if(value instanceof String){
                return Boolean.valueOf((String)value);
            } else {
                return false;
            }
        }

        public Float floatValue(){
            if(value==null){
                return (float)-1;
            } else if(value instanceof Float){
                return (Float)value;
            } else if(value instanceof String){
                return Float.valueOf((String)value);
            } else {
                return (float)-1;
            }
        }

        public String toString(){
            return value.toString();
        }
    }

    /**
     * Example usage.
     * @param args
     */
    public static void main(String[] args){
        IPSClient client=null;
        try {
            client=new IPSClient("192.168.1.23", 3774);
            client.logon("tf", "kosch");
            //client.pingServer();

            List<Var> vars=client.listVariables();

            for(Var var : vars){
                client.getValue(var);
                System.out.println(var.getName()+" "+var.toString());
            }

            Var var=new Var("FS1.Status");
            var.setType(Var.Type.BOOLEAN);
            var.setValue(true);
            client.setValue(var);

        } catch (Exception ex){
            ex.printStackTrace();
        } finally{
            if(client!=null){
                try {
                    client.logoff();
                } catch (Throwable t){}
            }
        }
    }

}
